"""Create and save plots for SIMD-Accelerated Distance computations."""

from __future__ import annotations

import time
import typing

import abd_distances.strings as string_distances
import editdistance
import tqdm

from . import utils

Functions = tuple[
    typing.Callable[[str, str], int],
    tuple[typing.Optional[str], typing.Optional[typing.Callable[[str, str], int]]],
]

FUNCTIONS: dict[str, Functions] = {
    "Hamming": (string_distances.hamming, (None, None)),
    "Levenshtein": (string_distances.levenshtein, ("roy-ht", editdistance.eval)),
    "Needleman-Wunsch": (string_distances.needleman_wunsch, (None, None)),
}


def make_plots() -> None:  # noqa: C901
    """Plot speed of string distance computations."""
    alphabet = "ACTGN"

    for name, (abd_func, (comp_name, comp_func)) in tqdm.tqdm(
        FUNCTIONS.items(),
        desc="String distances",
    ):
        x: list[int] = []
        y_abd: list[float] = []

        y_comp: list[float] | None = None
        if comp_name:
            y_comp = []

        car = 10
        str_lengths = list(range(1, 9)) + list(range(10, 51, 5))
        str_lengths = [i * 100 for i in str_lengths]
        for dim in str_lengths:
            x.append(dim)

            data = utils.gen_strings(car, dim, alphabet)

            start = time.perf_counter()
            for a in data:
                for b in data:
                    _ = abd_func(a, b)
            end = float(time.perf_counter() - start) * 1000
            if "hamming" in name.lower():
                end *= 1000
            end = end / (car**2)
            y_abd.append(end)

            if y_comp is not None:
                start = time.perf_counter()
                for a in data:
                    for b in data:
                        _ = comp_func(a, b)
                end_comp = float(time.perf_counter() - start) * 1000
                if "hamming" in name.lower():
                    end_comp *= 1000
                end_comp = end_comp / (car**2)
                y_comp.append(end_comp)

        y_units = "µs" if "hamming" in name.lower() else "ms"
        y_comp_label = comp_name if y_comp else ""

        utils.make_plot(
            x=x,
            y_abd=y_abd,
            fn_name=name,
            dt_name="str",
            y_units=y_units,
            x_label="String Length",
            y_comp=y_comp,
            y_comp_label=y_comp_label,
        )



"""A package for computing distances."""

from . import simd
from . import strings
from . import vectors

setattr(vectors, "cityblock", vectors.manhattan)  # noqa: B010

__all__ = [
    "simd",
    "strings",
    "vectors",
]

__version__ = "1.0.4"



"""Create and save plots for SIMD-Accelerated Distance computations."""

import time
import typing

import abd_distances.simd as simd_distances
import abd_distances.vectors as vector_distances
import numpy
import scipy.spatial.distance as scipy_distance
import tqdm

from . import utils

Functions = tuple[
    typing.Callable[[numpy.ndarray, numpy.ndarray], float],
    typing.Callable[[numpy.ndarray, numpy.ndarray], float],
    typing.Callable[[int, int], numpy.ndarray],
]


FUNCTION_PAIRS: dict[str, Functions] = {
    "SIMD Euclidean, f32": (
        simd_distances.euclidean,
        scipy_distance.euclidean,
        utils.data_f32,
    ),
    "SIMD Euclidean, f64": (
        simd_distances.euclidean,
        scipy_distance.euclidean,
        utils.data_f64,
    ),
    "SIMD Squared Euclidean, f32": (
        simd_distances.sqeuclidean,
        scipy_distance.sqeuclidean,
        utils.data_f32,
    ),
    "SIMD Squared Euclidean, f64": (
        simd_distances.sqeuclidean,
        scipy_distance.sqeuclidean,
        utils.data_f64,
    ),
    "SIMD Cosine, f32": (simd_distances.cosine, scipy_distance.cosine, utils.data_f32),
    "SIMD Cosine, f64": (simd_distances.cosine, scipy_distance.cosine, utils.data_f64),
    "Chebyshev, f32": (
        vector_distances.chebyshev,
        scipy_distance.chebyshev,
        utils.data_f32,
    ),
    "Chebyshev, f64": (
        vector_distances.chebyshev,
        scipy_distance.chebyshev,
        utils.data_f64,
    ),
    "Euclidean, f32": (
        vector_distances.euclidean,
        scipy_distance.euclidean,
        utils.data_f32,
    ),
    "Euclidean, f64": (
        vector_distances.euclidean,
        scipy_distance.euclidean,
        utils.data_f64,
    ),
    "Squared Euclidean, f32": (
        vector_distances.sqeuclidean,
        scipy_distance.sqeuclidean,
        utils.data_f32,
    ),
    "Squared Euclidean, f64": (
        vector_distances.sqeuclidean,
        scipy_distance.sqeuclidean,
        utils.data_f64,
    ),
    "Manhattan, f32": (
        vector_distances.manhattan,
        scipy_distance.cityblock,
        utils.data_f32,
    ),
    "Manhattan, f64": (
        vector_distances.manhattan,
        scipy_distance.cityblock,
        utils.data_f64,
    ),
    "Canberra, f32": (
        vector_distances.canberra,
        scipy_distance.canberra,
        utils.data_f32,
    ),  # type: ignore[attr-defined]
    "Canberra, f64": (
        vector_distances.canberra,
        scipy_distance.canberra,
        utils.data_f64,
    ),  # type: ignore[attr-defined]
    "Cosine, f32": (vector_distances.cosine, scipy_distance.cosine, utils.data_f32),
    "Cosine, f64": (vector_distances.cosine, scipy_distance.cosine, utils.data_f64),
    "Bray-Curtis, u32": (
        vector_distances.braycurtis,
        scipy_distance.braycurtis,
        utils.data_u32,
    ),  # type: ignore[attr-defined]
    "Bray-Curtis, u64": (
        vector_distances.braycurtis,
        scipy_distance.braycurtis,
        utils.data_u64,
    ),  # type: ignore[attr-defined]
    # "Hamming, i32": (vector_distances.hamming, scipy_distance.hamming, utils.data_i32),  # type: ignore[attr-defined]  # noqa: E501
    # "Hamming, i64": (vector_distances.hamming, scipy_distance.hamming, utils.data_i64),  # type: ignore[attr-defined]  # noqa: E501
}


def make_plots() -> None:
    """Plot SIMD speedup for various distance computations."""
    for name, (abd_func, scipy_func, gen_data) in tqdm.tqdm(
        FUNCTION_PAIRS.items(),
        desc="Vector Distances",
    ):
        [dist_name, dtype] = name.split(", ")
        dist_name = "-".join(dist_name.split(" "))

        x: list[int] = []
        y_abd: list[float] = []
        y_scipy: list[float] = []

        car = 100
        for d in range(5, 51, 5):
            dim = d * 100
            data = gen_data(car, dim)

            start = time.perf_counter_ns()
            for a in data:
                for b in data:
                    _ = abd_func(a, b)
            end_abd = float(time.perf_counter_ns() - start)
            end_abd = end_abd / (car**2) / 1000

            start = time.perf_counter_ns()
            for a in data:
                for b in data:
                    _ = scipy_func(a, b)
            end_scipy = float(time.perf_counter_ns() - start)
            end_scipy = end_scipy / (car**2) / 1000

            x.append(dim)
            y_abd.append(end_abd)
            y_scipy.append(end_scipy)

        utils.make_plot(
            x=x,
            y_abd=y_abd,
            y_comp=y_scipy,
            fn_name=dist_name,
            dt_name=dtype,
        )



"""Pytest configuration file."""

import numpy


def data_f32(car: int, dim: int) -> numpy.ndarray:
    """Return a random array of the given dimension."""
    rng = numpy.random.default_rng()
    return rng.random((car, dim)).astype(numpy.float32)


def data_f64(car: int, dim: int) -> numpy.ndarray:
    """Return a random array of the given dimension."""
    rng = numpy.random.default_rng()
    return rng.random((car, dim)).astype(numpy.float64)


def data_i32(car: int, dim: int) -> numpy.ndarray:
    """Return a random array of the given dimension."""
    rng = numpy.random.default_rng()
    data = rng.integers(0, 2, (car, dim)).astype(numpy.int32)
    data[data == 0] = -1
    return data


def data_i64(car: int, dim: int) -> numpy.ndarray:
    """Return a random array of the given dimension."""
    rng = numpy.random.default_rng()
    data = rng.integers(0, 2, (car, dim)).astype(numpy.int64)
    data[data == 0] = -1
    return data


def data_u32(car: int, dim: int) -> numpy.ndarray:
    """Return a random array of the given dimension."""
    rng = numpy.random.default_rng()
    return rng.integers(0, 2, (car, dim)).astype(numpy.uint32)


def data_u64(car: int, dim: int) -> numpy.ndarray:
    """Return a random array of the given dimension."""
    rng = numpy.random.default_rng()
    return rng.integers(0, 2, (car, dim)).astype(numpy.uint64)



"""Tests for the SIMD-accelerated distance functions."""

import math
import typing

import numpy
import scipy.spatial.distance as scipy_distance
from abd_distances import simd as simd_distances

Functions = tuple[
    typing.Callable[[numpy.ndarray, numpy.ndarray], float],
    typing.Callable[[numpy.ndarray, numpy.ndarray], float],
]

METRICS = ["euclidean", "sqeuclidean", "cosine"]


FUNCTIONS: dict[str, Functions] = {
    "Euclidean": (simd_distances.euclidean, scipy_distance.euclidean),
    "SqEuclidean": (simd_distances.sqeuclidean, scipy_distance.sqeuclidean),
    "Cosine": (simd_distances.cosine, scipy_distance.cosine),
}


def test_simd_f32(data_f32: numpy.ndarray):
    """Test the SIMD-accelerated distance functions."""
    for a in data_f32:
        for b in data_f32:
            for a_, b_ in [(a, a), (a, b), (b, a), (b, b)]:
                for name, (simd_func, scipy_func) in FUNCTIONS.items():
                    _simd_helper(a_, b_, name, simd_func, scipy_func, 1e-5)


def test_simd_f64(data_f64: numpy.ndarray):
    """Test the SIMD-accelerated distance functions."""
    for a in data_f64:
        for b in data_f64:
            for a_, b_ in [(a, a), (a, b), (b, a), (b, b)]:
                for name, (simd_func, scipy_func) in FUNCTIONS.items():
                    _simd_helper(a_, b_, name, simd_func, scipy_func, 1e-10)


def test_cdist_f32(data_f32: numpy.ndarray):
    """Test the SIMD-accelerated distance functions."""
    for metric in METRICS:
        _cdist_helper(
            data_f32,
            data_f32,
            metric,
            simd_distances.cdist,
            1e-5,
        )


def test_cdist_f64(data_f64: numpy.ndarray):
    """Test the SIMD-accelerated distance functions."""
    for metric in METRICS:
        _cdist_helper(
            data_f64,
            data_f64,
            metric,
            simd_distances.cdist,
            1e-10,
        )


def test_pdist_f32(data_f32: numpy.ndarray):
    """Test the SIMD-accelerated distance functions."""
    for metric in METRICS:
        _pdist_helper(
            data_f32,
            metric,
            simd_distances.pdist,
            1e-5,
        )


def test_pdist_f64(data_f64: numpy.ndarray):
    """Test the SIMD-accelerated distance functions."""
    for metric in METRICS:
        _pdist_helper(
            data_f64,
            metric,
            simd_distances.pdist,
            1e-10,
        )


def _simd_helper(  # noqa: PLR0913
    a: numpy.ndarray,
    b: numpy.ndarray,
    metric: str,
    simd_func: typing.Callable[[numpy.ndarray, numpy.ndarray], float],
    scipy_func: typing.Callable[[numpy.ndarray, numpy.ndarray], float],
    abs_tol: float,
) -> None:
    """Helper function for the SIMD-accelerated distance functions."""
    dist = simd_func(a, b)
    expected = scipy_func(a, b)
    abs_tol = abs_tol * expected if "cosine" not in metric.lower() else abs_tol
    assert math.isclose(
        dist,
        expected,
        abs_tol=abs_tol,
    ), f"{metric}: Expected: {expected:.8e}, got: {dist:.8e}, rel_tol: {abs_tol:.8e}"


def _cdist_helper(
    a: numpy.ndarray,
    b: numpy.ndarray,
    metric: str,
    simd_func: typing.Callable[[numpy.ndarray, numpy.ndarray, str], numpy.ndarray],
    abs_tol: float,
) -> None:
    """Helper function for the SIMD-accelerated distance functions."""
    distances = simd_func(a, b, metric)
    assert distances.shape == (a.shape[0], b.shape[0])
    expected = scipy_distance.cdist(a, b, metric)
    abs_tol = abs_tol * expected if "cosine" not in metric.lower() else abs_tol
    assert numpy.allclose(
        distances,
        expected,
        atol=abs_tol,
    ), f"{metric}: Expected: {expected}, got: {distances}, rel_tol: {abs_tol}"


def _pdist_helper(
    a: numpy.ndarray,
    metric: str,
    simd_func: typing.Callable[[numpy.ndarray, str], numpy.ndarray],
    abs_tol: float,
) -> None:
    """Helper function for the SIMD-accelerated distance functions."""
    distances = simd_func(a, metric)
    num_distances = (a.shape[0] * (a.shape[0] - 1)) // 2
    assert distances.shape == (num_distances,)
    expected = scipy_distance.pdist(a, metric)
    abs_tol = abs_tol * expected if "cosine" not in metric.lower() else abs_tol
    assert numpy.allclose(
        distances,
        expected,
        atol=abs_tol,
    ), f"{metric}: Expected: {expected}, got: {distances}, rel_tol: {abs_tol}"



def hamming(a: str, b: str) -> int:
    """Compute the Hamming distance between two strings."""

def levenshtein(a: str, b: str) -> int:
    """Compute the Levenshtein distance between two strings."""

def needleman_wunsch(a: str, b: str) -> int:
    """Compute the Needleman-Wunsch distance between two strings."""



"""CLI for making plots."""

from . import plot_strings
from . import plot_vectors


def main() -> None:
    """Run the CLI."""
    plot_strings.make_plots()
    plot_vectors.make_plots()


if __name__ == "__main__":
    main()



"""Benchmark vector distances."""

import typing
from functools import partial

import numpy
import scipy.spatial.distance as scipy_distance
import utils  # type: ignore[import]
from abd_distances import simd

CAR = 20
DIM = 500
NUM_RUNS = 100

GEN_DATA = {
    "f32": utils.data_f32,
    "f64": utils.data_f64,
}

METRICS = [
    "euclidean",
    "sqeuclidean",
    "cosine",
]

FUNCTIONS = {
    f"{name}, {dtype}": (
        getattr(scipy_distance, name),
        getattr(simd, name),
        gen_data,
    )
    for name in METRICS
    for dtype, gen_data in GEN_DATA.items()
}


def bench_func(
    func: typing.Callable[[numpy.ndarray, numpy.ndarray], float],
    gen_data: typing.Callable[[int, int], numpy.ndarray],
) -> None:
    """Benchmark a distance function."""
    data_x = gen_data(CAR, DIM)
    data_y = gen_data(CAR, DIM)
    for _ in range(NUM_RUNS):
        for a in data_x:
            for b in data_y:
                _ = func(a, b)


__benchmarks__ = [
    (
        partial(bench_func, scipy_func, gen_data),
        partial(bench_func, abd_func, gen_data),
        f"SIMD, {name}",
    )
    for name, (scipy_func, abd_func, gen_data) in FUNCTIONS.items()
]
